/*entry.h
  this file defines class Entry*/
#ifndef __Entry 
#define __Entry
//------------------------------------------------------------
#include "../func/gendef.h"
//------------------------------------------------------------
class RTNode;
class RTree;
struct Linkable;
//------------------------------------------------------------
class Entry 
{
public:
	int b_blk;
	int dimension;                      // dimension of the box
	int level;                     
	  //note level field of entry is not stored in the disk.  its main usage
	  //is in the insertion function.  so before inserting any entry, remember 
	  //to set this field.
	int son;
	float *bounces;                     // pointer to box
	RTree *my_tree;                     // pointer to my R-tree
	  //Tips: for classes that inherit this one, overload this variable.  i.e.
	  //define a my_tree which point to another type (say MVRTree).
    RTNode *son_ptr;                    // pointer to son if in main mem.
	  //overload this variable in case of inheritance

	//--====functions===--
	Entry();
	Entry(int dimension, RTree *rt);
    ~Entry();

	void del_son();
	Linkable *gen_Linkable();
	int get_size();                     // returns amount of needed buffer space
	RTNode *get_son();                  // returns pointer to the son,
                                        // loads son if necessary
	void init_entry(int _dimension, RTree *_rt);
	void read_from_buffer(char *buffer);// reads data from buffer
    SECTION section(float *mbr);        // tests, if mbr intersects the box
	void set_from_Linkable(Linkable *link);
    bool section_circle(float *center, float radius);
    void write_to_buffer(char *buffer); // writes data to buffer

	bool operator == (Entry &_d);
	//-----------------------------------------------------
	//-----the following functions are modified for aRB-tree implementation---
	//-----------------------------------------------------
	virtual Entry & operator = (Entry &_d);
	//-----------------------------------------------------
	//-----the following functions are added for aRB-tree implementation---
	//-----------------------------------------------------
	void insert_agg(int _tm, int _agg);
};

#endif