/* rtree.cpp this file implements the RTree class */
#include <math.h>
#include <string.h>
#include "rtree.h"
#include "entry.h"
#include "rtnode.h"
#include "../blockfile/cache.h"
#include "../blockfile/blk_file.h"
#include "../linlist/linlist.h"
#include "../heap/heap.h"
//------------------------------------------------------------
RTree::RTree(char *_fname, Cache *_cache)
  //use this constructor to restore a tree from a file
{
	//---initiate the cacheable class---
	char rfname[50];
	strcpy(rfname, _fname);
	strcat(rfname, ".rb");
    file = new BlockFile(rfname, 0);
    cache = _cache;
	//----------------------------------

	//---initiate the member variables---
    re_data_cands = new LinList();     //re-insert link list
	char *header = new char [file -> get_blocklength()];
    file -> read_header(header);
    read_header(header);
	delete [] header;	
	root_ptr = NULL;
	//----------------------------------

	bitmaplen=hashnum*bitlen;
	bitcharlen=bitmaplen/8;

	hp=new Heap();
	hp->init(10000, bitcharlen);

}

/*****************************************************************
this constructor builds the host tree in the arb-tree
it takes the data file in the format of <id,x0,x1,y0,y1>
para:
dfname: the data file name
tfname: the tree file name
b_len: the block length
cache: 
dim: the dimensionality
hashnum: how many hash functions used
bitlen: how many bits are used for the signature of each function
pcsa: toggles if pcsa is used in the signatures
coded by yufei tao, may 2003
******************************************************************/

RTree::RTree(char *_dfname, char *_tfname, int _b_len, Cache *_cache, int _dim, int _hashnum, int _bitlen,
			 bool _pcsa)
{
	if (_bitlen%8)
		error("bitlen must be a multiple of 8, qutting from rtree constructor...\n", true);

	//---initiate the cacheable class---
	char rfname[50];
	strcpy(rfname, _tfname);
	strcat(rfname, ".rb");
    file = new BlockFile(rfname, _b_len);
    cache = _cache;
	//----------------------------------

	//---initiate the member variables---
    re_data_cands = new LinList();     //re-insert link list
	dimension = _dim;
    root_is_data = TRUE;
    num_of_data = num_of_inodes = num_of_dnodes = 0;
	fin_time = -1;
	hashnum=_hashnum; bitlen=_bitlen;
	bitmaplen=hashnum*bitlen;
	bitcharlen=bitmaplen/8;
	pcsa=_pcsa;
	hp=NULL;
	//----------------------------------

	//---create the root (with no entries) of the r-tree---
    root_ptr = new RTNode(this);
    root_ptr -> level = 0;
    root = root_ptr->block;
	num_of_dnodes++;
	//----------------------------------

	FILE *fp;

	int rec_cnt = 0;     //record count
    if((fp = fopen(_dfname,"r")) == NULL)
    {
		delete this;
		error("Cannot open R-Tree text file", TRUE);
    }
    else
    {
		while (!feof(fp))
		{
			rec_cnt ++;
			if (rec_cnt % 1000 == 0)
			{
				for (int i = 0; i < 79; i ++)  printf("\b");
				  //clear a line
					printf("inserting object %d", rec_cnt);
			}

			int id, agg;
			float x0, x1, y0, y1;
			fscanf(fp, "%d %f %f %f %f\n", &id, &x0, &x1, &y0, &y1);

			Entry *d = new Entry(dimension, NULL);
			d -> son = id;
			d -> bounces[0] = x0;
			d -> bounces[1] = x1;
			d -> bounces[2] = y0;
			d -> bounces[3] = y1;

			insert(d);
			  //d will be deleted in insert()
		}
	}
	fclose(fp);
	printf("\n");
	del_root();
}

//------------------------------------------------------------

RTree::~RTree()
{
	char *header = new char[file -> get_blocklength()];
    write_header(header);
    file -> set_header(header);
    delete [] header;

    if (root_ptr != NULL)
    {
        delete root_ptr;
        root_ptr = NULL;
    }

	if (cache) cache -> flush();

    delete file;
    delete re_data_cands;

	if (hp)
		delete hp;

    printf("This R-Tree contains %d internal, %d data nodes and %d data\n",
	   num_of_inodes, num_of_dnodes, num_of_data);
}

//------------------------------------------------------------

void RTree::del_root()
{
	delete root_ptr; root_ptr = NULL;
}

//------------------------------------------------------------

bool RTree::findLeaf(Entry *e)
{
	load_root();
	return root_ptr -> FindLeaf(e);
	del_root();
}

//------------------------------------------------------------

void RTree::insert(Entry* d)
{
    int i, j;
    RTNode *sn;
    RTNode *nroot_ptr;
    int nroot;
    Entry *de;
    R_OVERFLOW split_root;
    Entry *dc;
    float *nmbr;

    // load root into memory
    load_root();

    // no overflow occured until now
    re_level = new bool[root_ptr -> level + 1];
    for (i = 0; i <= root_ptr -> level; i++)
        re_level[i] = FALSE;

    // insert d into re_data_cands as the first entry to insert
    // make a copy of d because it should be erased later
    Linkable *new_link;
	new_link = d -> gen_Linkable();
	re_data_cands -> insert(new_link);

	delete d;  //we follow the convention that the entry will be deleted when insertion finishes

    j = -1;
    while (re_data_cands -> get_num() > 0)
    {
        // first try to insert data, then directory entries
	    Linkable *d_cand;
		d_cand = re_data_cands -> get_first();
        if (d_cand != NULL)
        {
            // since "erase" deletes the data itself from the
            // list, we should make a copy of the data before
            // erasing it
			dc = new Entry(dimension, NULL);
            dc -> set_from_Linkable(d_cand);
            re_data_cands -> erase();

            // start recursive insert with root
			split_root = root_ptr -> insert(dc, &sn);
        }
        else
	        error("RTree::insert: inconsistent list re_data_cands", TRUE);

    	if (split_root == SPLIT)
    	// insert has lead to split --> new root-page with two sons (i.e. root and sn)
    	{
    	    nroot_ptr = new RTNode(this);
    	    nroot_ptr -> level = root_ptr -> level + 1;
    	    num_of_inodes++;
    	    nroot = nroot_ptr -> block;

    	    de = new Entry(dimension, this);
    	    nmbr = root_ptr -> get_mbr();
    	    memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
    	    delete [] nmbr;
    	    de->son = root_ptr->block;
    	    de->son_ptr = root_ptr;
    	    nroot_ptr -> enter(de);

    	    de = new Entry(dimension, this);
    	    nmbr = sn -> get_mbr();
    	    memcpy(de -> bounces, nmbr, 2*dimension*sizeof(float));
    	    delete [] nmbr;
    	    de -> son = sn -> block;
    	    de -> son_ptr = sn;
    	    nroot_ptr->enter(de);

    	    root = nroot;
            root_ptr = nroot_ptr;

            root_is_data = FALSE;
        }
        j++;
    }

    num_of_data++;

    delete [] re_level;

	delete root_ptr;
	root_ptr = NULL;
}

//------------------------------------------------------------

void RTree::load_root()
{
    if (root_ptr == NULL)
        root_ptr = new RTNode(this, root);
}

//------------------------------------------------------------

void RTree::rangeQuery(float *mbr, SortedLinList *res)
{
    load_root();

    root_ptr -> rangeQuery(mbr,res);

	delete root_ptr;
	root_ptr = NULL;
}

//------------------------------------------------------------

void RTree::read_header(char *buffer)
{
    int i;

    memcpy(&dimension, buffer, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&num_of_data, &buffer[i], sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&num_of_dnodes, &buffer[i], sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&num_of_inodes, &buffer[i], sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&root_is_data, &buffer[i], sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&root, &buffer[i], sizeof(root));
    i += sizeof(root);

	memcpy(&fin_time, &buffer[i], sizeof(fin_time));
	i += sizeof(fin_time);

	memcpy(&hashnum, &buffer[i], sizeof(hashnum));
	i+=sizeof(hashnum);

	memcpy(&bitlen, &buffer[i], sizeof(bitlen));
	i+=sizeof(bitlen);

	memcpy(&pcsa, &buffer[i], sizeof(pcsa));
	i+=sizeof(pcsa);
}

//------------------------------------------------------------

void RTree::write_header(char *buffer)
{
    int i;

    memcpy(buffer, &dimension, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&buffer[i], &num_of_data, sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&buffer[i], &num_of_dnodes, sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&buffer[i], &num_of_inodes, sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&buffer[i], &root_is_data, sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&buffer[i], &root, sizeof(root));
    i += sizeof(root);

	memcpy(&buffer[i], &fin_time, sizeof(fin_time));
	i+=sizeof(fin_time);

	memcpy(&buffer[i], &hashnum, sizeof(hashnum));
	i+=sizeof(hashnum);

	memcpy(&buffer[i], &bitlen, sizeof(bitlen));
	i+=sizeof(bitlen);

	memcpy(&buffer[i], &pcsa, sizeof(pcsa));
	i+=sizeof(pcsa);
}

/*****************************************************************
this function creates the b-trees associated with the host entries. 
para:
dfname: the spatial file name, containing the mbrs of the rectangles
sigfname: the signature file name
rnum: number of regions
NOTE: regions' ids are assumed to be within [0, rnum-1]
coded by yufei tao, may 2003
******************************************************************/

void RTree::create_btrees(char *_dfname, char *_sigfname, int _rnum)
{
	FILE *dfp=fopen(_dfname, "r");
	if (!dfp)
		error("could not open the spatial file...\n", true);
	FILE *sfp=fopen(_sigfname, "rb");
	if (!sfp)
		error("could not open the signature file...\n", true);

	printf("reading the spatial dataset...\n");
	float *reg=new float[4*_rnum];
	int cnt=0;
	while (!feof(dfp))
	{
		if (cnt==_rnum)
			error("\tthe spatial dataset's cardinality is larger than you thought...\n", true);
		int id; 
		fscanf(dfp, "%d ", &id);
		for (int i=0; i<4; i++)
			fscanf(dfp, "%f ", &reg[4*cnt+i]);
		fscanf(dfp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt!=_rnum)
		printf("\twarning: the spatial dataset's cardinality is smaller than you thought...\n");

	printf("inserting bitmaps...\n");
	unsigned char *bitchars=new unsigned char[bitcharlen];
	unsigned char *thisbitchars=new unsigned char[bitcharlen];
	unsigned char *zerobitchars=new unsigned char[bitcharlen];
	for (int i=0; i<bitcharlen; i++)
		zerobitchars[i]=0;

	bool *reginst=new bool[_rnum];
	for (i=0; i<_rnum; i++)
		reginst[i]=false;
	int last_tm=-1; bool first=true;
	int tm, rid;
	binary_int_read(sfp, &tm);
	while(tm!=-1)
	{
		binary_int_read(sfp, &rid);

		for (int i=0; i<bitcharlen; i++)
			bitchars[i]=fgetc(sfp);

		if (tm>last_tm)
		{
			if (first)
				first=false;
			else
			{
				for (int i=0; i<_rnum; i++)
					if (!reginst[i])
						insert_bitmap(last_tm, i, &reg[i*4], zerobitchars);
				for (i=0; i<_rnum; i++)
					reginst[i]=false;
			}
			last_tm=tm;
			printf("\ttimestamp %d...\n", tm);
		}

//if (tm==1 && rid==9)
//printf("testing...\n");

		insert_bitmap(tm,rid,&reg[rid*4],bitchars);
		reginst[rid]=true;
//added to check the integrity----
/*
int ctm=1, crid=2;
get_reg_bitchars(crid, &reg[crid*4], ctm, thisbitchars);
for (i=0; i<bitcharlen; i++)
if (thisbitchars[i]!=bitchars[i])
{
	printf("tm=%d, rid=%d\n", tm, rid);
	error("caught a bug in rtree::create_btrees\n", true);
}
*/
//--------------------------------
		binary_int_read(sfp, &tm);
	}

	delete []reginst;
	delete []zerobitchars;
	delete []thisbitchars;
	delete []bitchars;
	delete []reg;

	fclose(sfp);
	fclose(dfp);
}

/*****************************************************************
this function inserts the char presentation of a region's bitmap. 
para:
tm: the timestamp of the insertion
rid: the region id
rmbr: the mbr of the region
bitchar: the char representation of the bitmap
coded by yufei tao, may 2003
******************************************************************/

void RTree::insert_bitmap(int _tm, int _rid, float *_rmbr, unsigned char *_bitchar)
{
	load_root();
	int cret=root_ptr->insert_bitmap(_tm, _rid, _rmbr, _bitchar);
	if (!cret)
	{
		printf("id=%d, <%.3f, %.3f, %.3f, %.3f>\n", _rid, _rmbr[0], _rmbr[1], _rmbr[2], _rmbr[3]);
		error("\tError: the required region not found...\n", true);
	}
	del_root();
}

/*****************************************************************
this function performs an approximate aggregate query on this arb-tree
para:
qmbr: the spatial query mbr
qt: the query interval
ed:	please give the last timestamp of the history
coded by yufei tao, may 2003
******************************************************************/

float RTree::agg_qry(float *_qmbr, float *_qt, float _ed)
{
//debugging---
//qfp=fopen("accessed.txt", "w");
//------------
	bool *qbitmaps=new bool[bitmaplen];
	for (int i=0; i<bitmaplen; i++)
		qbitmaps[i]=false;
	load_root();
	root_ptr->agg_qry(qbitmaps, _qmbr, _qt, _ed);
	del_root();

	//now perform estimate
	int total=0;
	for (i=0; i<hashnum; i++)
	{
		int pos=0;
		for (int j=0; j<bitlen; j++)
			if (!qbitmaps[i*bitlen+j])
			{
				pos=j; break;
			}
		total+=pos;
	}
	float tmp=pow(2.0, ((float)total)/hashnum);
	float est;
	if (pcsa)
		est=(hashnum/0.77351)*tmp;
	else
		est=tmp/0.77351;
	delete []qbitmaps;
//debugging---
//fclose(qfp);
//------------
	return est;
}

/*****************************************************************
this function gets the bitchar of a single region at some timestamp
para:
rid: the region id
rmbr: the region's mbr
tm: the timestamp
bitchar: the char representation of the bitmap
coded by yufei tao, may 2003
******************************************************************/

void RTree::get_reg_bitchars(int _rid, float *_rmbr, int _tm, unsigned char *_bitchar)
{
	load_root();
	bool cret=root_ptr->get_reg_bitchars(_rid, _rmbr, _tm, _bitchar);
	if (!cret)
		error("the region not found\n", true);
	del_root();
}

/*****************************************************************
this function performs an approximate aggregate query on this arb-tree
it implements a better algorithm with heap
para:
qmbr: the spatial query mbr
qt: the query interval
ed:	please give the last timestamp of the history
coded by yufei tao, june 2003
******************************************************************/

//auxiliary function
void RTree::conservative_pos_est(bool *_qbitmaps, Heap *_hp, int *_pos)
{
	bool *or=new bool[bitmaplen];
	for (int i=0; i<bitmaplen; i++)
		or[i]=_qbitmaps[i];
	bool *bitmap=new bool[bitmaplen];
	for (i=0; i<_hp->used; i++)
	{
		char2bitmap(bitmap, _hp->cont[i].bitchars, bitmaplen);
		for (int j=0; j<bitmaplen; j++)
			or[j]=or[j]|bitmap[j];
	}

	for (i=0; i<hashnum; i++)
		for (int j=0; j<bitlen; j++)
			if (!or[i*bitlen+j])
			{	
				_pos[i]=j; break;
			}

	delete []bitmap;
	delete []or;
}

//auxiliary function
bool RTree::visit_child(bool *_qbitmaps, unsigned char *_bitchars, int *_pos)
{
	bool ret=false;
	bool *bitmaps=new bool[bitmaplen];
	char2bitmap(bitmaps, _bitchars, bitmaplen);
	for (int i=0; i<hashnum; i++)
	{
		for (int j=0; j<_pos[i]; j++)
//		for (int j=0; j<bitlen; j++)
			if ((bitmaps[i*bitlen+j]|_qbitmaps[i*bitlen+j])!=_qbitmaps[i*bitlen+j])
			{
				ret=true; 
				break; 
			}
		if (ret)
			break;
	}

	delete []bitmaps;
	return ret;
}

float RTree::agg_qry2(float *_qmbr, float *_qt, float _ed)
{
//debugging---
//qfp=fopen("accessed.txt", "w");
//------------
	bool *qbitmaps=new bool[bitmaplen];
	for (int i=0; i<bitmaplen; i++)
		qbitmaps[i]=false;

	hp->used=0;
	load_heap(qbitmaps, _qmbr, _qt, _ed);
	
	int *pos=new int[hashnum];

	HeapEntry *he=new HeapEntry();
	he->init_HeapEntry(bitcharlen);
	while (hp->used>0)
	{
		conservative_pos_est(qbitmaps, hp, pos);
		hp->remove(he);
		if (visit_child(qbitmaps, he->bitchars, pos))
		{
			int son=he->son;
//debugging---
//fprintf(qfp, "%d\n", son);
//------------
			BNode *bn=new BNode(this, son);

			for (int i=0; i<bn->num_entries; i++)
			{
				int st=bn->entries[i].tm, ed;
				if (i!=bn->num_entries-1)
					ed=bn->entries[i+1].tm;
				else
					ed=he->ed;
				if (bn->level==0)
				{
					if (max(st,_qt[0])<=min(ed-1,_qt[1]))
					{
						bool *bitmaps=new bool[bitmaplen];
						char2bitmap(bitmaps, bn->entries[i].bitchars, bitmaplen);
						for (int j=0; j<bitmaplen; j++)
							qbitmaps[j]|=bitmaps[j];
						delete []bitmaps;
					}
				}
				else
				{
					if ((_qt[0]<=st) && (ed-1<=_qt[1]))
					{
						bool *bitmaps=new bool[bitmaplen];
						char2bitmap(bitmaps, bn->entries[i].bitchars, bitmaplen);
						for (int j=0; j<bitmaplen; j++)
							qbitmaps[j]|=bitmaps[j];
						delete []bitmaps;
					}
					else if (max(st,_qt[0])<=min(ed-1,_qt[1]))
					{
						memcpy(he->bitchars, bn->entries[i].bitchars, bitcharlen);
						he->level=bn->level;
						he->son=bn->entries[i].son;
						he->key=get_one_number(he->bitchars, bitmaplen);
if (he->son==96231)
printf("");
						hp->insert(he);
					}
				}
			}
			delete bn;
		}
	}
	delete he;

	//now perform estimate
	int total=0;
	for (i=0; i<hashnum; i++)
	{
		int pos=0;
		for (int j=0; j<bitlen; j++)
			if (!qbitmaps[i*bitlen+j])
			{
				pos=j; break;
			}
		total+=pos;
	}
	float tmp=pow(2.0, ((float)total)/hashnum);
	float est;
	if (pcsa)
		est=(hashnum/0.77351)*tmp;
	else
		est=tmp/0.77351;

	delete []pos;
	delete []qbitmaps;
//debugging---
//fclose(qfp);
//------------
	return est;
}

/*****************************************************************
this function loads the heap (for query) with the qualifying entries
in the roots of the B-trees
para:
qmbr: the spatial query mbr
qt: the query interval
ed:	please give the last timestamp of the history
coded by yufei tao, june 2003
******************************************************************/

void RTree::load_heap(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed)
{
	load_root();
	root_ptr->load_heap(_qbitmaps, _qmbr, _qt, _ed);
	del_root();
}